@@ -1,23 +1,20 @@
 /*
- * Copyright (C) 2014 Glyptodon LLC
+ * Licensed to the Apache Software Foundation (ASF) under one
+ * or more contributor license agreements.  See the NOTICE file
+ * distributed with this work for additional information
+ * regarding copyright ownership.  The ASF licenses this file
+ * to you under the Apache License, Version 2.0 (the
+ * "License"); you may not use this file except in compliance
+ * with the License.  You may obtain a copy of the License at
  *
- * Permission is hereby granted, free of charge, to any person obtaining a copy
- * of this software and associated documentation files (the "Software"), to deal
- * in the Software without restriction, including without limitation the rights
- * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
- * copies of the Software, and to permit persons to whom the Software is
- * furnished to do so, subject to the following conditions:
+ *   http://www.apache.org/licenses/LICENSE-2.0
  *
- * The above copyright notice and this permission notice shall be included in
- * all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
- * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
- * THE SOFTWARE.
+ * Unless required by applicable law or agreed to in writing,
+ * software distributed under the License is distributed on an
+ * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ * KIND, either express or implied.  See the License for the
+ * specific language governing permissions and limitations
+ * under the License.
  */
 
 /**
@@ -35,9 +32,13 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
     // Required services
     var $location             = $injector.get('$location');
     var authenticationService = $injector.get('authenticationService');
+    var clipboardService      = $injector.get('clipboardService');
     var guacClientManager     = $injector.get('guacClientManager');
     var guacNotification      = $injector.get('guacNotification');
+    var iconService           = $injector.get('iconService');
     var preferenceService     = $injector.get('preferenceService');
+    var requestService        = $injector.get('requestService');
+    var tunnelService         = $injector.get('tunnelService');
     var userPageService       = $injector.get('userPageService');
 
     /**
@@ -64,7 +65,7 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
      */
     var MENU_DRAG_VERTICAL_TOLERANCE = 10;
 
-    /*
+    /**
      * In order to open the guacamole menu, we need to hit ctrl-alt-shift. There are
      * several possible keysysms for each key.
      */
@@ -74,6 +75,20 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         CTRL_KEYS   = {0xFFE3 : true, 0xFFE4 : true},
         MENU_KEYS   = angular.extend({}, SHIFT_KEYS, ALT_KEYS, CTRL_KEYS);
 
+    /**
+     * Keysym for detecting any END key presses, for the purpose of passing through
+     * the Ctrl-Alt-Del sequence to a remote system.
+     */
+    var END_KEYS = {0xFF57 : true, 0xFFB1 : true};
+
+    /**
+     * Keysym for sending the DELETE key when the Ctrl-Alt-End hotkey
+     * combo is pressed.
+     *
+     * @type Number
+     */
+    var DEL_KEY = 0xFFFF;
+
     /**
      * All client error codes handled and passed off for translation. Any error
      * code not present in this list will be represented by the "DEFAULT"
@@ -83,7 +98,11 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         0x0201: true,
         0x0202: true,
         0x0203: true,
-        0x0205: true,
+        0x0207: true,
+        0x0208: true,
+        0x0209: true,
+        0x020A: true,
+        0x020B: true,
         0x0301: true,
         0x0303: true,
         0x0308: true,
@@ -98,6 +117,8 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         0x0200: true,
         0x0202: true,
         0x0203: true,
+        0x0207: true,
+        0x0208: true,
         0x0301: true,
         0x0308: true
     };
@@ -113,6 +134,8 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         0x0203: true,
         0x0204: true,
         0x0205: true,
+        0x0207: true,
+        0x0208: true,
         0x0301: true,
         0x0303: true,
         0x0308: true,
@@ -127,6 +150,8 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         0x0200: true,
         0x0202: true,
         0x0203: true,
+        0x0207: true,
+        0x0208: true,
         0x0308: true
     };
 
@@ -137,7 +162,9 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         name      : "CLIENT.ACTION_LOGOUT",
         className : "logout button",
         callback  : function logoutCallback() {
-            authenticationService.logout()['finally'](function logoutComplete() {
+            authenticationService.logout()
+            ['catch'](requestService.IGNORE)
+            ['finally'](function logoutComplete() {
                 $location.url('/');
             });
         }
@@ -164,7 +191,7 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
             };
         }
 
-    });
+    }, requestService.WARN);
 
     /**
      * Action which replaces the current client with a newly-connected client.
@@ -230,11 +257,6 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         $scope.menu.shown = false;
     };
 
-    // Update the model when clipboard data received from client
-    $scope.$on('guacClientClipboard', function clientClipboardListener(event, client, mimetype, clipboardData) {
-       $scope.clipboardData = clipboardData; 
-    });
-
     /**
      * The client which should be attached to the client UI.
      *
@@ -242,8 +264,43 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
      */
     $scope.client = guacClientManager.getManagedClient($routeParams.id, $routeParams.params);
 
+    /**
+     * Map of all available sharing profiles for the current connection by
+     * their identifiers. If this information is not yet available, or no such
+     * sharing profiles exist, this will be an empty object.
+     *
+     * @type Object.<String, SharingProfile>
+     */
+    $scope.sharingProfiles = {};
+
+    /**
+     * Map of all currently pressed keys by keysym. If a particular key is
+     * currently pressed, the value stored under that key's keysym within this
+     * map will be true. All keys not currently pressed will not have entries
+     * within this map.
+     *
+     * @type Object.<Number, Boolean>
+     */
     var keysCurrentlyPressed = {};
 
+    /**
+     * Map of all substituted key presses.  If one key is pressed in place of another
+     * the value of the substituted key is stored in an object with the keysym of
+     * the original key.
+     *
+     * @type Object.<Number, Number>
+     */
+    var substituteKeysPressed = {};
+
+    /**
+     * Map of all currently pressed keys (by keysym) to the clipboard contents
+     * received from the remote desktop while those keys were pressed. All keys
+     * not currently pressed will not have entries within this map.
+     *
+     * @type Object.<Number, ClipboardData>
+     */
+    var clipboardDataFromKey = {};
+
     /*
      * Check to see if all currently pressed keys are in the set of menu keys.
      */  
@@ -376,27 +433,105 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         
         // Send clipboard data if menu is hidden
         if (!menuShown && menuShownPreviousState)
-            $scope.$broadcast('guacClipboard', 'text/plain', $scope.client.clipboardData); 
+            $scope.$broadcast('guacClipboard', $scope.client.clipboardData);
         
         // Disable client keyboard if the menu is shown
         $scope.client.clientProperties.keyboardEnabled = !menuShown;
 
     });
-    
+
+    // Update page icon when thumbnail changes
+    $scope.$watch('client.thumbnail.canvas', function thumbnailChanged(canvas) {
+        iconService.setIcons(canvas);
+    });
+
+    // Watch clipboard for new data, associating it with any pressed keys
+    $scope.$watch('client.clipboardData', function clipboardChanged(data) {
+
+        // Sync local clipboard as long as the menu is not open
+        if (!$scope.menu.shown)
+            clipboardService.setLocalClipboard(data)['catch'](angular.noop);
+
+        // Associate new clipboard data with any currently-pressed key
+        for (var keysym in keysCurrentlyPressed)
+            clipboardDataFromKey[keysym] = data;
+
+    });
+
+    // Pull sharing profiles once the tunnel UUID is known
+    $scope.$watch('client.tunnel.uuid', function retrieveSharingProfiles(uuid) {
+
+        // Only pull sharing profiles if tunnel UUID is actually available
+        if (!uuid)
+            return;
+
+        // Pull sharing profiles for the current connection
+        tunnelService.getSharingProfiles(uuid)
+        .then(function sharingProfilesRetrieved(sharingProfiles) {
+            $scope.sharingProfiles = sharingProfiles;
+        }, requestService.WARN);
+
+    });
+
+    /**
+     * Produces a sharing link for the current connection using the given
+     * sharing profile. The resulting sharing link, and any required login
+     * information, will be displayed to the user within the Guacamole menu.
+     *
+     * @param {SharingProfile} sharingProfile
+     *     The sharing profile to use to generate the sharing link.
+     */
+    $scope.share = function share(sharingProfile) {
+        ManagedClient.createShareLink($scope.client, sharingProfile);
+    };
+
+    /**
+     * Returns whether the current connection has any associated share links.
+     *
+     * @returns {Boolean}
+     *     true if the current connection has at least one associated share
+     *     link, false otherwise.
+     */
+    $scope.isShared = function isShared() {
+        return ManagedClient.isShared($scope.client);
+    };
+
+    /**
+     * Returns the total number of share links associated with the current
+     * connection.
+     *
+     * @returns {Number}
+     *     The total number of share links associated with the current
+     *     connection.
+     */
+    $scope.getShareLinkCount = function getShareLinkCount() {
+
+        // Count total number of links within the ManagedClient's share link map
+        var linkCount = 0;
+        for (var dummy in $scope.client.shareLinks)
+            linkCount++;
+
+        return linkCount;
+
+    };
+
+    // Track pressed keys, opening the Guacamole menu after Ctrl+Alt+Shift, or
+    // send Ctrl-Alt-Delete when Ctrl-Alt-End is pressed.
     $scope.$on('guacKeydown', function keydownListener(event, keysym, keyboard) {
+
+        // Record key as pressed
         keysCurrentlyPressed[keysym] = true;   
         
-        /* 
-         * If only menu keys are pressed, and we have one keysym from each group,
-         * and one of the keys is being released, show the menu. 
-         */
-        if(checkMenuModeActive()) {
-            var currentKeysPressedKeys = Object.keys(keysCurrentlyPressed);
+        var currentKeysPressedKeys = Object.keys(keysCurrentlyPressed);
+
+        // If only menu keys are pressed, and we have one keysym from each group,
+        // and one of the keys is being released, show the menu. 
+        if (checkMenuModeActive()) {
             
             // Check that there is a key pressed for each of the required key classes
-            if(!_.isEmpty(_.pick(SHIFT_KEYS, currentKeysPressedKeys)) &&
-               !_.isEmpty(_.pick(ALT_KEYS, currentKeysPressedKeys)) &&
-               !_.isEmpty(_.pick(CTRL_KEYS, currentKeysPressedKeys))
+            if (!_.isEmpty(_.pick(SHIFT_KEYS, currentKeysPressedKeys)) &&
+                !_.isEmpty(_.pick(ALT_KEYS, currentKeysPressedKeys)) &&
+                !_.isEmpty(_.pick(CTRL_KEYS, currentKeysPressedKeys))
             ) {
         
                 // Don't send this key event through to the client
@@ -412,16 +547,58 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
                 });
             }
         }
+
+        // If one of the End keys is pressed, and we have a one keysym from each
+        // of Ctrl and Alt groups, send Ctrl-Alt-Delete.
+        if (END_KEYS[keysym] &&
+            !_.isEmpty(_.pick(ALT_KEYS, currentKeysPressedKeys)) &&
+            !_.isEmpty(_.pick(CTRL_KEYS, currentKeysPressedKeys))
+        ) {
+
+            // Don't send this event through to the client.
+            event.preventDefault();
+
+            // Remove the original key press
+            delete keysCurrentlyPressed[keysym];
+
+            // Record the substituted key press so that it can be
+            // properly dealt with later.
+            substituteKeysPressed[keysym] = DEL_KEY;
+
+            // Send through the delete key.
+            $scope.$broadcast('guacSyntheticKeydown', DEL_KEY);
+        }
+
     });
 
-    // Listen for broadcasted keyup events and fire the appropriate listeners
+    // Update pressed keys as they are released, synchronizing the clipboard
+    // with any data that appears to have come from those key presses
     $scope.$on('guacKeyup', function keyupListener(event, keysym, keyboard) {
-        delete keysCurrentlyPressed[keysym];
+
+        // Sync local clipboard with any clipboard data received while this
+        // key was pressed (if any) as long as the menu is not open
+        var clipboardData = clipboardDataFromKey[keysym];
+        if (clipboardData && !$scope.menu.shown)
+            clipboardService.setLocalClipboard(clipboardData)['catch'](angular.noop);
+
+        // Deal with substitute key presses
+        if (substituteKeysPressed[keysym]) {
+            event.preventDefault();
+            delete substituteKeysPressed[keysym];
+            $scope.$broadcast('guacSyntheticKeyup', substituteKeysPressed[keysym]);
+        }
+
+        // Mark key as released
+        else {
+            delete clipboardDataFromKey[keysym];
+            delete keysCurrentlyPressed[keysym];
+        }
+
     });
 
-    // Update page title when client name is received
-    $scope.$watch('client.name', function clientNameChanged(name) {
-        $scope.page.title = name;
+    // Update page title when client title changes
+    $scope.$watch('client.title', function clientTitleChanged(title) {
+        $scope.page.title = title;
     });
 
     /**
@@ -440,6 +617,7 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
 
         // Re-authenticate to verify auth status at end of connection
         authenticationService.updateCurrentToken($location.search())
+        ['catch'](requestService.IGNORE)
 
         // Show the requested status once the authentication check has finished
         ['finally'](function authenticationCheckComplete() {
@@ -448,6 +626,18 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
 
     };
 
+    /**
+     * Returns whether the current connection has been flagged as unstable due
+     * to an apparent network disruption.
+     *
+     * @returns {Boolean}
+     *     true if the current connection has been flagged as unstable, false
+     *     otherwise.
+     */
+    $scope.isConnectionUnstable = function isConnectionUnstable() {
+        return $scope.client && $scope.client.clientState.tunnelUnstable;
+    };
+
     // Show status dialog when connection status changes
     $scope.$watch('client.clientState.connectionState', function clientStateChanged(connectionState) {
 
@@ -473,7 +663,9 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
          || connectionState === ManagedClientState.ConnectionState.WAITING) {
             guacNotification.showStatus({
                 title: "CLIENT.DIALOG_HEADER_CONNECTING",
-                text: "CLIENT.TEXT_CLIENT_STATUS_" + connectionState.toUpperCase()
+                text: {
+                    key : "CLIENT.TEXT_CLIENT_STATUS_" + connectionState.toUpperCase()
+                }
             });
         }
 
@@ -490,7 +682,9 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
             notifyConnectionClosed({
                 className : "error",
                 title     : "CLIENT.DIALOG_HEADER_CONNECTION_ERROR",
-                text      : "CLIENT.ERROR_CLIENT_" + errorName,
+                text      : {
+                    key : "CLIENT.ERROR_CLIENT_" + errorName
+                },
                 countdown : countdown,
                 actions   : actions
             });
@@ -510,7 +704,9 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
             notifyConnectionClosed({
                 className : "error",
                 title     : "CLIENT.DIALOG_HEADER_CONNECTION_ERROR",
-                text      : "CLIENT.ERROR_TUNNEL_" + errorName,
+                text      : {
+                    key : "CLIENT.ERROR_TUNNEL_" + errorName
+                },
                 countdown : countdown,
                 actions   : actions
             });
@@ -521,21 +717,32 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         else if (connectionState === ManagedClientState.ConnectionState.DISCONNECTED) {
             notifyConnectionClosed({
                 title   : "CLIENT.DIALOG_HEADER_DISCONNECTED",
-                text    : "CLIENT.TEXT_CLIENT_STATUS_" + connectionState.toUpperCase(),
+                text    : {
+                    key : "CLIENT.TEXT_CLIENT_STATUS_" + connectionState.toUpperCase()
+                },
                 actions : actions
             });
         }
 
+        // Hide status and sync local clipboard once connected
+        else if (connectionState === ManagedClientState.ConnectionState.CONNECTED) {
+
+            // Sync with local clipboard
+            clipboardService.getLocalClipboard().then(function clipboardRead(data) {
+                $scope.$broadcast('guacClipboard', data);
+            }, angular.noop);
+
+            // Hide status notification
+            guacNotification.showStatus(false);
+
+        }
+
         // Hide status for all other states
         else
             guacNotification.showStatus(false);
 
     });
 
-    $scope.formattedScale = function formattedScale() {
-        return Math.round($scope.client.clientProperties.scale * 100);
-    };
-    
     $scope.zoomIn = function zoomIn() {
         $scope.menu.autoFit = false;
         $scope.client.clientProperties.autoFit = false;
@@ -546,6 +753,16 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         $scope.client.clientProperties.autoFit = false;
         $scope.client.clientProperties.scale -= 0.1;
     };
+
+    /**
+     * When zoom is manually set by entering a value
+     * into the controller, this method turns off autoFit,
+     * both in the menu and the clientProperties.
+     */
+    $scope.zoomSet = function zoomSet() {
+        $scope.menu.autoFit = false;
+        $scope.client.clientProperties.autoFit = false;
+    };
     
     $scope.changeAutoFit = function changeAutoFit() {
         if ($scope.menu.autoFit && $scope.client.clientProperties.minScale) {
@@ -707,7 +924,27 @@ angular.module('client').controller('clientController', ['$scope', '$routeParams
         if (!$scope.client)
             return false;
 
-        return !!($scope.client.uploads.length || $scope.client.downloads.length);
+        return !!$scope.client.uploads.length;
+
+    };
+
+    /**
+     * Returns whether the current user can share the current connection with
+     * other users. A connection can be shared if and only if there is at least
+     * one associated sharing profile.
+     *
+     * @returns {Boolean}
+     *     true if the current user can share the current connection with other
+     *     users, false otherwise.
+     */
+    $scope.canShareConnection = function canShareConnection() {
+
+        // If there is at least one sharing profile, the connection can be shared
+        for (var dummy in $scope.sharingProfiles)
+            return true;
+
+        // Otherwise, sharing is not possible
+        return false;
 
     };
 
